/*
 * Ben NanoNote board initialisation, based on uboot-xburst and xburst-tools.
 *
 * Copyright (C) 2015 Paul Boddie <paul@boddie.org.uk>
 * Copyright (C) Xiangfu Liu <xiangfu.z@gmail.com>
 * Copyright (C) 2006 Ingenic Semiconductor, <jlwei@ingenic.cn>
 * Copyright (C) 2000-2009 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "jz4740.h"
#include "nanonote.h"
#include "usb_boot_defines.h"

/* These arguments are initialised by usbboot and are defined in...
   /etc/xburst-tools/usbboot.cfg. */

struct fw_args *fw_args;
volatile u32 FW_CPU_ID;
volatile u8 FW_SDRAM_BW16;
volatile u8 FW_SDRAM_BANK4;
volatile u8 FW_SDRAM_ROW;
volatile u8 FW_SDRAM_COL;
volatile u8 FW_CONFIG_MOBILE_SDRAM;
volatile u8 FW_IS_SHARE;

void load_args(void)
{
	/* Get the fw args from memory. See head1.S for the memory layout. */

        fw_args = (struct fw_args *)0x80002008;
        FW_CPU_ID = fw_args->cpu_id ;
        FW_SDRAM_BW16 = fw_args->bus_width;
        FW_SDRAM_BANK4 = fw_args->bank_num;
        FW_SDRAM_ROW = fw_args->row_addr;
        FW_SDRAM_COL = fw_args->col_addr;
        FW_CONFIG_MOBILE_SDRAM = fw_args->is_mobile;
        FW_IS_SHARE = fw_args->is_busshare;
}

/* Initialisation functions. */

void gpio_init(void)
{
	/*
	 * Initialize NAND Flash Pins
	 */
	__gpio_as_nand();

	/*
	 * Initialize SDRAM pins
	 */
	__gpio_as_sdram_16bit_4720();
}

void pll_init(void)
{
	register unsigned int cfcr, plcr1;
	int nf, pllout2;

	/* See CPCCR (Clock Control Register).
	 * 0 == same frequency; 2 == f/3
	 */

	cfcr = CPM_CPCCR_CLKOEN |
		CPM_CPCCR_PCS |
		(0 << CPM_CPCCR_CDIV_BIT) |
		(2 << CPM_CPCCR_HDIV_BIT) |
		(2 << CPM_CPCCR_PDIV_BIT) |
		(2 << CPM_CPCCR_MDIV_BIT) |
		(2 << CPM_CPCCR_LDIV_BIT);

	/* Determine the divider clock output based on the PCS bit. */

	pllout2 = (cfcr & CPM_CPCCR_PCS) ? CONFIG_SYS_CPU_SPEED : (CONFIG_SYS_CPU_SPEED / 2);

	/* Init USB Host clock.
	 * Divisor == UHCCDR + 1
	 * Desired frequency == 48MHz
	 */

	REG_CPM_UHCCDR = pllout2 / 48000000 - 1;

	nf = CONFIG_SYS_CPU_SPEED * 2 / CONFIG_SYS_EXTAL;
	plcr1 = ((nf - 2) << CPM_CPPCR_PLLM_BIT) |	/* FD */
		(0 << CPM_CPPCR_PLLN_BIT) |		/* RD=0, NR=2 */
		(0 << CPM_CPPCR_PLLOD_BIT) |		/* OD=0, NO=1 */
		CPM_CPPCR_PLLEN;			/* enable PLL */

	/* Update PLL and wait. */

	REG_CPM_CPCCR = cfcr;
	REG_CPM_CPPCR = plcr1;
	while (!__cpm_pll_is_on());
}

void sdram_init(void)
{
	register unsigned int dmcr0, dmcr, sdmode, tmp, cpu_clk, mem_clk, ns;
	unsigned int pllout = __cpm_get_pllout();

	unsigned int cas_latency_sdmr[2] = {
		EMC_SDMR_CAS_2,
		EMC_SDMR_CAS_3,
	};

	unsigned int cas_latency_dmcr[2] = {
		1 << EMC_DMCR_TCL_BIT,	/* CAS latency is 2 */
		2 << EMC_DMCR_TCL_BIT	/* CAS latency is 3 */
	};

	/* Divisors for CPCCR values. */

	int div[] = {1, 2, 3, 4, 6, 8, 12, 16, 24, 32};

	cpu_clk = pllout / div[__cpm_get_cdiv()];
	mem_clk = pllout / div[__cpm_get_mdiv()];

	REG_EMC_BCR = 0;	/* Disable bus release */
	REG_EMC_RTCSR = 0;	/* Disable clock for counting */

	/* Fault DMCR value for mode register setting*/
#define SDRAM_ROW0    11
#define SDRAM_COL0     8
#define SDRAM_BANK40   0

	dmcr0 = ((SDRAM_ROW0-11)<<EMC_DMCR_RA_BIT) |
		((SDRAM_COL0-8)<<EMC_DMCR_CA_BIT) |
		(SDRAM_BANK40<<EMC_DMCR_BA_BIT) |
		(FW_SDRAM_BW16<<EMC_DMCR_BW_BIT) |
		EMC_DMCR_EPIN |
		cas_latency_dmcr[((SDRAM_CASL == 3) ? 1 : 0)];

	/* Basic DMCR value */
	dmcr = ((FW_SDRAM_ROW-11)<<EMC_DMCR_RA_BIT) |
		((FW_SDRAM_COL-8)<<EMC_DMCR_CA_BIT) |
		(FW_SDRAM_BANK4<<EMC_DMCR_BA_BIT) |
		(FW_SDRAM_BW16<<EMC_DMCR_BW_BIT) |
		EMC_DMCR_EPIN |
		cas_latency_dmcr[((SDRAM_CASL == 3) ? 1 : 0)];

	/* SDRAM timimg */
	ns = 1000000000 / mem_clk;
	tmp = SDRAM_TRAS/ns;
	if (tmp < 4) tmp = 4;
	if (tmp > 11) tmp = 11;
	dmcr |= ((tmp-4) << EMC_DMCR_TRAS_BIT);
	tmp = SDRAM_RCD/ns;
	if (tmp > 3) tmp = 3;
	dmcr |= (tmp << EMC_DMCR_RCD_BIT);
	tmp = SDRAM_TPC/ns;
	if (tmp > 7) tmp = 7;
	dmcr |= (tmp << EMC_DMCR_TPC_BIT);
	tmp = SDRAM_TRWL/ns;
	if (tmp > 3) tmp = 3;
	dmcr |= (tmp << EMC_DMCR_TRWL_BIT);
	tmp = (SDRAM_TRAS + SDRAM_TPC)/ns;
	if (tmp > 14) tmp = 14;
	dmcr |= (((tmp + 1) >> 1) << EMC_DMCR_TRC_BIT);

	/* SDRAM mode value */
	sdmode = EMC_SDMR_BT_SEQ | 
		 EMC_SDMR_OM_NORMAL |
		 EMC_SDMR_BL_4 | 
		 cas_latency_sdmr[((SDRAM_CASL == 3) ? 1 : 0)];

	/* Stage 1. Precharge all banks by writing SDMR with DMCR.MRSET=0 */
	REG_EMC_DMCR = dmcr;
	REG8(EMC_SDMR0|sdmode) = 0;

	/* Wait for precharge, > 200us */
	tmp = (cpu_clk / 1000000) * 1000;
	while (tmp--);

	/* Stage 2. Enable auto-refresh */
	REG_EMC_DMCR = dmcr | EMC_DMCR_RFSH;

	tmp = SDRAM_TREF/ns;
	tmp = tmp/64 + 1;
	if (tmp > 0xff) tmp = 0xff;
	REG_EMC_RTCOR = tmp;
	REG_EMC_RTCNT = 0;
	REG_EMC_RTCSR = EMC_RTCSR_CKS_64;	/* Divisor is 64, CKO/64 */

	/* Wait for number of auto-refresh cycles */
	tmp = (cpu_clk / 1000000) * 1000;
	while (tmp--);

 	/* Stage 3. Mode Register Set */
	REG_EMC_DMCR = dmcr0 | EMC_DMCR_RFSH | EMC_DMCR_MRSET;
	REG8(EMC_SDMR0|sdmode) = 0;

        /* Set back to basic DMCR value */
	REG_EMC_DMCR = dmcr | EMC_DMCR_RFSH | EMC_DMCR_MRSET;

	/* everything is ok now */
}
